import os
os.environ["OMP_NUM_THREADS"] = "1"
import numpy as np
import scipy.special as ss
from collections import defaultdict,OrderedDict
import matplotlib.pyplot as plt
#from MainLib import ToMemb
import sys
import pandas as pd

def ToMemb(nod,N):
    h = np.array([-1]*N)

    nod = sorted(nod,key=lambda x:len(x),reverse=True)
    Hn = []
    sel = {}
    while True:
        c = 0
        for i in range(len(nod)):
            if i in sel: continue
            if not all(h[np.array(nod[i])]==-1): continue

            h[np.array(nod[i])]=c
            sel[i] = None
            c+=1

        h[h==-1] = np.arange(h.max()+1,(h==-1).sum()+h.max()+1)
        Hn.append(h)
        h = np.array([-1]*N)
        if len(sel)==len(nod):
            break

    Hn.append(np.arange(N))    
    return Hn

    
def StandardDendrogram(H):
	h = sorted(list(zip(*H)))
	h = list(map(list,h))

	for j in range(len(h[0])-1):
		k = OrderedDict.fromkeys(list(zip(*h))[j]).keys()
		k = dict(zip(k,range(len(k))))

		for i in range(len(h)):
			h[i][j] = k[h[i][j]]

	j = len(h[0])-1

	k = OrderedDict.fromkeys(list(zip(*h))[j]).keys()
	k = dict(zip(k,range(len(k))))

	h = sorted(h,key=lambda x:x[-1])

	for i in range(len(h)):
			h[i][j] = k[h[i][j]]
			
	h = list(map(np.array,zip(*h)))
	return h

def OrderArray(H):
    return np.array( list(zip(*sorted(zip(H[-1],range(len(H[0]))))))[1] )


def FiltMatrix(R,clust,method='average'):
	N = R.shape[0]

	Rv = np.identity(N)

	Done = ~np.identity(N).astype(bool)

	C = map(np.array,sorted(clust,key=len))

	for c in C:
		one = np.zeros(N)

		one[c] = 1
		if method=='average':
			Rv[np.where(np.outer(one,one)*Done)] = R[np.where(np.outer(one,one)*Done)].mean()
		elif method=='complete':
			 Rv[np.where(np.outer(one,one)*Done)] = R[np.where(np.outer(one,one)*Done)].min()
		elif method=='single':
			 Rv[np.where(np.outer(one,one)*Done)] = R[np.where(np.outer(one,one)*Done)].max()


		Done[np.where(np.outer(one,one)*Done)] = False

	return Rv



import os
os.environ["OMP_NUM_THREADS"] = "1"

import numpy as np
import fastcluster
import pandas as pd
from joblib import Parallel, delayed
import scipy.stats as st
from functools import partial
import itertools
from multiprocessing import cpu_count
import meboot

def flatten(container):
    for i in container:
        if isinstance(i, (list,tuple)):
            for j in flatten(i):
                yield j
        else:
            yield i
def flatx(x):
    return tuple(flatten(x))


def dist(R,method):
    N = R.shape[0]
    d = R[np.triu_indices(N,1)]

    if method=='average':
        out = fastcluster.average(d)
    if method=='complete':
        out = fastcluster.complete(d)
    if method=='single':
        out = fastcluster.single(d)
  
    outI = out.astype(int)
    dend = {i:(i,) for i in range(N)}

    for i in range(len(outI)):
        dend[i+N] = (dend[outI[i][0]],dend[outI[i][1]])

    for i in range(N):
        dend.pop(i,None)

    dend = [(flatx(a),flatx(b)) for a,b in dend.values()]

    dend ={flatx((a,b)):(np.array(a),np.array(b)) for a,b in dend}
    
    return dend


def singPV(Rb,method):
    
    if method=='average':
        rxy = dict(map(lambda x: (x[0],Rb[x[1][0]][:,x[1][1]].mean()),LV.items()))
    if method=='complete':
        rxy = dict(map(lambda x: (x[0],Rb[x[1][0]][:,x[1][1]].max()),LV.items()))
    if method=='single':
        rxy = dict(map(lambda x: (x[0],Rb[x[1][0]][:,x[1][1]].min()),LV.items()))

    PV = list(map(lambda x:int(rxy[x[1]]<rxy[x[0]]), gen.items()))
    
    return PV



def SingleBoot(nan,method,metric,bootstrap,sel):

	if nan==False and bootstrap=='simple':
		Xb = Xg[:,sel]
	elif nan==False and bootstrap=='meboot':
		Xb = meb.bootstrap()

	if nan==False:
		if metric=='pearson':
			Rb = 1.-np.corrcoef(Xb)
		elif metric=='spearman':
			Rb = 1.-st.spearmanr(Xb.T)[0]
	else:
		Rb = 1. -Xg[sel].T.corr(method=metric).values

	return singPV(Rb,method)


def BootDist(method,Nt=1000,nan=False,ncpu=1,metric='pearson',bootstrap='simple'):

	global gen
	gen = {tuple(s):k for k in LV for s in LV[k] if len(s)>1}

	if bootstrap=='simple':
		sels = np.random.choice(range(Xg.shape[1]),replace=True,size=(Nt,Xg.shape[1]))
	elif bootstrap=='meboot':
		sels = range(Nt)
		global meb
		meb = meboot.MEBOOT(Xg)
	 
	f = partial(SingleBoot,nan,method,metric,bootstrap)

	PV = Parallel(n_jobs=ncpu, backend="threading")(delayed(f)(sel) for sel in sels)

	return sorted(zip(np.array(PV).mean(axis=0),gen.keys()))
		
def FDR(PV,alpha,N):
    p = np.array(list(zip(*PV))[0])
    thr = np.arange(1,len(PV)+1)*alpha/len(PV)

    sel = np.where(p<thr)[0]
    if len(sel)==0: 
        thr=-1.
    else:
        thr = thr[sel][-1]

    L = [range(N)]+[PV[i][1] for i in np.where(p<=thr)[0]]
    L = list(map(tuple,sorted(L,key=len,reverse=True)))

    PV = {c:p for p,c in PV}
    PV[tuple(range(N))] = np.nan
    return L,PV

def Bootstrap_PV(LV,X,Nt):
    gen = {tuple(s):k for k in LV for s in LV[k] if len(s)>1}

    sels = sd.choice(range(X.shape[1]),replace=False,size=(Nt,X.shape[1]))

    PV = Pvalues(X,sels,LV,gen)
    return PV

def Analytic_PV(r,n):
    R = 1-r
    gen = {tuple(s):k for k in LV for s in LV[k] if len(s)>1}

    sd = np.vectorize(lambda x: (1./n)*(1-x**2)**2)
    COV = lambda i,j: (0.5/n)*(((R[i[0],j[0]]-R[i[0],i[1]]*R[j[0],i[1]])*(R[i[1],j[1]]-R[i[1],j[0]]*R[j[0],j[1]]))+((R[i[0],j[1]]-R[i[0],j[0]]*R[j[0],j[1]])*(R[i[1],j[0]]-R[i[1],i[0]]*R[i[0],j[0]]))+((R[i[0],j[0]]-R[i[0],j[1]]*R[j[1],j[0]])*(R[i[1],j[1]]-R[i[1],i[0]]*R[i[0],j[1]]))+((R[i[0],j[1]]-R[i[0],i[1]]*R[i[1],j[1]])*(R[i[1],j[0]]-R[i[1],j[1]]*R[j[1],j[0]])))
    PV = []
    for h,(son,fat) in enumerate(gen.items()):


        a_s,b_s = LV[son]
        a_f,b_f = LV[fat]

        m = R[a_s][:,b_s].mean()-R[a_f][:,b_f].mean()

        ss = sd(R[a_s][:,b_s])
        sf = sd(R[a_f][:,b_f])

        elm_f = list(itertools.product(a_f, b_f))
        elm_s = list(itertools.product(a_s, b_s))

        ns = 1./np.size(ss)
        nf = 1./np.size(sf)

        cf = (nf**2)*2*sum(COV(elm_f[i],elm_f[j]) for i in range(len(elm_f)) for j in range(i))
        cs = (ns**2)*2*sum(COV(elm_s[i],elm_s[j]) for i in range(len(elm_s)) for j in range(i))
        cmix = -2*(ns*nf)*sum(COV(elm_f[i],elm_s[j]) for i in range(len(elm_f)) for j in range(len(elm_s)))

        x = (ns**2)*ss.sum()+(nf**2)*sf.sum()+cf+cs+cmix

        s = np.sqrt(x)

        PV.append((st.norm.cdf(0,m,s),son))

    return PV

def Find_ValidatedCluster(X,Nt=1000,alpha=0.05,ncpu=cpu_count(),dendrogram=None,method='average',metric='pearson',bootstrap='simple'):

	global sd
	sd = np.random.RandomState(0)

	nan = np.isnan(X).any()
	if nan==True and bootstrap=='meboot':
		raise Exception('meeboot does not accept nan')

	if nan==False:
		if metric=='pearson':
			R = 1. - np.corrcoef(X)
		elif metric=='spearman':
			R = 1. -st.spearmanr(X.T)[0]
	else:
		X = pd.DataFrame(X)
		R = 1. -X.T.corr(method=metric).values

	global Xg
	Xg = X

	global LV
	if dendrogram==None:
		LV = dist(R,method)
	else:
		LV = dendrogram


	if Nt==0:
		PV = Analytic_PV(R,X.shape[1])
	else:
		PV = BootDist(method,Nt,nan,ncpu,metric,bootstrap)

	PV = [(p,tuple(sorted(k))) for p,k in PV]

	L,PV = FDR(PV,alpha,X.shape[0])

	return {'validated':L,'pvalues':PV,'dendrogram':LV}




import os
os.environ["OMP_NUM_THREADS"] = "1"
import matplotlib.pyplot as plt
import matplotlib.cm as cmap
from metric import StandardDendrogram,OrderArray,ToMemb
import matplotlib.patches as patches
import numpy as np
import seaborn as sns
import pandas as pd
import scipy.stats as st

def DendroAndCorrDist(x,validated,dendrogram,method='average',xylabel='Objects',file_w=None,color_style='Set1',nan=False,metric='pearson'):
    

    ordn = OrderArray(StandardDendrogram(ToMemb(dendrogram.keys(),x.shape[0])))
    x = x[ordn]

    rev_ornd = dict(sorted(zip(ordn,range(len(ordn)))))

    dendrogram = {tuple(sorted([rev_ornd[k] for k in K])):(np.array(sorted([rev_ornd[a] for a in dendrogram[K][0]])),
                                                   np.array(sorted([rev_ornd[a] for a in dendrogram[K][1]]))) for K in dendrogram}

    validated = [tuple(sorted([rev_ornd[k] for k in K])) for K in validated]

    if nan==False:
        if metric=='pearson':
            R = 1 - np.corrcoef(x)
        elif metric=='spearman':
            R = 1 - st.spearmanr(x.T)[0]
    else:
        R = 1. -np.array(pd.DataFrame(x.T).corr()) 
    #R = 1.-np.corrcoef(x)

    validated = sorted(validated,key=len)

    N = R.shape[0]
    K = dendrogram.keys()

    import matplotlib.patches as patches
    import colorsys
    nc  = len(validated)
    import matplotlib as mpl
    import matplotlib.cm as cm

    cnmx = sns.color_palette(color_style, nc-1)

    from collections import defaultdict
    col = defaultdict(int)
    for k in K:
        for j in range(len(validated)):
            if set(k) <= set(validated[j]):

                if tuple(validated[j])==tuple(range(N)):
                    col[tuple(k)] = 'black'
                    break

                col[tuple(k)] = cnmx[j-1]
                break


    from matplotlib import gridspec

    fig = plt.figure(figsize=(5,9))
    gs = gridspec.GridSpec(2, 1, height_ratios=[1, 1.34]) 

    # the fisrt subplot
    axarr = [0,0]
    axarr[0] = plt.subplot(gs[0])
    axarr[1] = plt.subplot(gs[1], sharex = axarr[0])


    pos = defaultdict(tuple)

    for i in range(N):
        pos[(i,)] = (float(i),0)

    K = sorted(dendrogram.keys(),key=len)

    hm = []
    for k in K:
        d = tuple(sorted(dendrogram[k][0])),tuple(sorted(dendrogram[k][1]))

        c = tuple(sorted(k))
        #if not col.has_key(c):
        #    break
        pp = pos[d[0]][0],pos[d[1]][0]
        h0 = pos[d[0]][1],pos[d[1]][1]

        if method=='average':
            h = R[dendrogram[k][0]][:,dendrogram[k][1]].mean()
        elif method=='complete':
            h = R[dendrogram[k][0]][:,dendrogram[k][1]].max()
        elif method=='single':
            h = R[dendrogram[k][0]][:,dendrogram[k][1]].min()

        axarr[0].plot([pp[0],pp[0]],[h0[0],h],'-',color='k',alpha=0.5)
        axarr[0].plot([pp[1],pp[1]],[h0[1],h],'-',color='k',alpha=0.5)
        axarr[0].plot([pp[0],pp[1]],[h,h],'-',color='k',lw=1.5)
        axarr[0].plot([pp[0],pp[1]],[h,h],'-',color=col[c],lw=1.4,zorder=1000)

        hm.append(h)

        #plt.plot((pp[0]+pp[1])/2,h,'o',color=col[c],ms=4,zorder=1000)

        pos[c] = (pp[0]+pp[1])/2,h
        #print "ciao"

    axarr[0].set_yticks(np.arange(0,1.2,0.2))
    axarr[0].set_ylim([0,max(hm)+0.05])

    axarr[0].yaxis.tick_right()
    axarr[0].tick_params(labelsize=14)
    axarr[0].set_ylabel('Distance',fontsize=18)
    axarr[0].yaxis.set_label_position("right")

    if nan==False:
        R = np.corrcoef(x)
    else:
        R =  np.array(pd.DataFrame(x.T).corr()) 

    ims = axarr[1].imshow(R,vmin=-1,vmax=1,cmap=cm.RdBu_r,interpolation='nearest')


    cbar_ax = fig.add_axes([0.95, 0.133, 0.03, 0.4])
    cbr = fig.colorbar(ims, cax=cbar_ax)

    cbr.set_label('correlation',fontsize=18)
    axarr[1].set_xlabel(xylabel,fontsize=18)
    axarr[1].set_ylabel(xylabel,fontsize=18)
    axarr[1].tick_params(labelsize=14)

    axarr[1].set_ylim([0,N])

    for l in validated:
        if l == tuple(range(N)): continue

        axarr[1].add_patch(
            patches.Rectangle(
                (min(l)-0.5, min(l)-0.5),
                len(l),
                len(l),
                fill=False,
                lw=1.,
                color='green',	
                )
            )

    plt.setp(axarr[0].get_xticklabels(), visible=False)
    plt.subplots_adjust(hspace=.0)

    if file_w!=None:    
        plt.savefig(file_w+'.pdf',bbox_inches='tight')
    else:
        plt.show()


def DendroAll(x,validated,dendrogram,method='average',xylabel='Objects',file_w=None,color_style='Set1',nan=False,metric='pearson'):
    

    ordn = OrderArray(StandardDendrogram(ToMemb(dendrogram.keys(),x.shape[0])))
    x = x[ordn]

    rev_ornd = dict(sorted(zip(ordn,range(len(ordn)))))

    dendrogram = {tuple(sorted([rev_ornd[k] for k in K])):(np.array(sorted([rev_ornd[a] for a in dendrogram[K][0]])),
                                                   np.array(sorted([rev_ornd[a] for a in dendrogram[K][1]]))) for K in dendrogram}

    validated = [tuple(sorted([rev_ornd[k] for k in K])) for K in validated]

    if nan==False:
        if metric=='pearson':
            R = 1 - np.corrcoef(x)
        elif metric=='spearman':
            R = 1 - st.spearmanr(x.T)[0]
    else:
        R = 1. -np.array(pd.DataFrame(x.T).corr()) 
    #R = 1.-np.corrcoef(x)

    validated = sorted(validated,key=len)

    N = R.shape[0]
    K = dendrogram.keys()

    import matplotlib.patches as patches
    import colorsys
    nc  = len(validated)
    import matplotlib as mpl
    import matplotlib.cm as cm

    cnmx = sns.color_palette(color_style, nc-1)

    from collections import defaultdict
    col = defaultdict(int)
    
    for k in K:
        for j in range(len(validated)):
            if set(k) <= set(validated[j]):

                if tuple(validated[j])==tuple(range(N)):
                    col[tuple(k)] = 'black'
                    break

                col[tuple(k)] = 'black'
                break


    from matplotlib import gridspec

    fig = plt.figure(figsize=(5,9))
    gs = gridspec.GridSpec(2, 1, height_ratios=[1, 1.34]) 

    # the fisrt subplot
    axarr = [0,0]
    axarr[0] = plt.subplot(gs[0])
    axarr[1] = plt.subplot(gs[1], sharex = axarr[0])


    pos = defaultdict(tuple)

    for i in range(N):
        pos[(i,)] = (float(i),0)

    K = sorted(dendrogram.keys(),key=len)

    hm = []
    for k in K:
        d = tuple(sorted(dendrogram[k][0])),tuple(sorted(dendrogram[k][1]))

        c = tuple(sorted(k))
        #if not col.has_key(c):
        #    break
        pp = pos[d[0]][0],pos[d[1]][0]
        h0 = pos[d[0]][1],pos[d[1]][1]

        if method=='average':
            h = R[dendrogram[k][0]][:,dendrogram[k][1]].mean()
        elif method=='complete':
            h = R[dendrogram[k][0]][:,dendrogram[k][1]].max()
        elif method=='single':
            h = R[dendrogram[k][0]][:,dendrogram[k][1]].min()

        axarr[0].plot([pp[0],pp[0]],[h0[0],h],'-',color='k',alpha=0.5)
        axarr[0].plot([pp[1],pp[1]],[h0[1],h],'-',color='k',alpha=0.5)
        axarr[0].plot([pp[0],pp[1]],[h,h],'-',color='k',lw=1.5)
        axarr[0].plot([pp[0],pp[1]],[h,h],'-',color='k',lw=1.4,zorder=1000)

        hm.append(h)

        #plt.plot((pp[0]+pp[1])/2,h,'o',color=col[c],ms=4,zorder=1000)

        pos[c] = (pp[0]+pp[1])/2,h
        #print "ciao"

    axarr[0].set_yticks(np.arange(0,1.2,0.2))
    axarr[0].set_ylim([0,max(hm)+0.05])

    axarr[0].yaxis.tick_right()
    axarr[0].tick_params(labelsize=14)
    axarr[0].set_ylabel('Distance',fontsize=18)
    axarr[0].yaxis.set_label_position("right")

    if nan==False:
        R = np.corrcoef(x)
    else:
        R =  np.array(pd.DataFrame(x.T).corr()) 

    ims = axarr[1].imshow(R,vmin=-1,vmax=1,cmap=cm.RdBu_r,interpolation='nearest')


    cbar_ax = fig.add_axes([0.95, 0.133, 0.03, 0.4])
    cbr = fig.colorbar(ims, cax=cbar_ax)

    cbr.set_label('correlation',fontsize=18)
    axarr[1].set_xlabel(xylabel,fontsize=18)
    axarr[1].set_ylabel(xylabel,fontsize=18)
    axarr[1].tick_params(labelsize=14)

    axarr[1].set_ylim([0,N])

    for l in validated:
        if l == tuple(range(N)): continue

        axarr[1].add_patch(
            patches.Rectangle(
                (min(l)-0.5, min(l)-0.5),
                len(l),
                len(l),
                fill=False,
                lw=1.,
                color='black',	
                )
            )

    plt.setp(axarr[0].get_xticklabels(), visible=False)
    plt.subplots_adjust(hspace=.0)


    if file_w!=None:    
        plt.savefig(file_w+'.pdf',bbox_inches='tight')
    else:
        plt.show()

def OnlyBox(x,validated,dendrogram,method='average',posx=[0.95, 0.285, 0.03, 0.43],xylabel='Objects',file_w=None,color_style='Set1',nan=False,metric='pearson'):
    

    ordn = OrderArray(StandardDendrogram(ToMemb(dendrogram.keys(),x.shape[0])))
    x = x[ordn]

    rev_ornd = dict(sorted(zip(ordn,range(len(ordn)))))

    dendrogram = {tuple(sorted([rev_ornd[k] for k in K])):(np.array(sorted([rev_ornd[a] for a in dendrogram[K][0]])),
                                                   np.array(sorted([rev_ornd[a] for a in dendrogram[K][1]]))) for K in dendrogram}

    validated = [tuple(sorted([rev_ornd[k] for k in K])) for K in validated]

    if nan==False:
        if metric=='pearson':
            R = 1 - np.corrcoef(x)
        elif metric=='spearman':
            R = 1 - st.spearmanr(x.T)[0]
    else:
        R = 1. -np.array(pd.DataFrame(x.T).corr()) 
    #R = 1.-np.corrcoef(x)

    validated = sorted(validated,key=len)

    N = R.shape[0]
    K = dendrogram.keys()

    import matplotlib.patches as patches
    import colorsys
    nc  = len(validated)
    import matplotlib as mpl
    import matplotlib.cm as cm

    cnmx = sns.color_palette(color_style, nc-1)

    from collections import defaultdict
    col = defaultdict(int)
    for k in K:
        for j in range(len(validated)):
            if set(k) <= set(validated[j]):

                if tuple(validated[j])==tuple(range(N)):
                    col[tuple(k)] = 'black'
                    break

                col[tuple(k)] = 'black'
                break


    from matplotlib import gridspec


    fig,arr = plt.subplots(figsize=(5,9))
    #gs = gridspec.GridSpec(1, 1, height_ratios=[1, 1.34]) 

    # the fisrt subplot
    #axarr = [0]
    #axarr[0] = plt.subplot(gs[0])
    #axarr[1] = plt.subplot(gs[1], sharex = axarr[0])


    pos = defaultdict(tuple)

    for i in range(N):
        pos[(i,)] = (float(i),0)

    K = sorted(dendrogram.keys(),key=len)

    hm = []
    for k in K:
        d = tuple(sorted(dendrogram[k][0])),tuple(sorted(dendrogram[k][1]))

        c = tuple(sorted(k))
        #if not col.has_key(c):
        #    break
        pp = pos[d[0]][0],pos[d[1]][0]
        h0 = pos[d[0]][1],pos[d[1]][1]

        if method=='average':
            h = R[dendrogram[k][0]][:,dendrogram[k][1]].mean()
        elif method=='complete':
            h = R[dendrogram[k][0]][:,dendrogram[k][1]].max()
        elif method=='single':
            h = R[dendrogram[k][0]][:,dendrogram[k][1]].min()

        #axarr[0].plot([pp[0],pp[0]],[h0[0],h],'-',color='k',alpha=0.5)
        #axarr[0].plot([pp[1],pp[1]],[h0[1],h],'-',color='k',alpha=0.5)
        #axarr[0].plot([pp[0],pp[1]],[h,h],'-',color='k',lw=1.5)
        #axarr[0].plot([pp[0],pp[1]],[h,h],'-',color='k',lw=1.4,zorder=1000)

        hm.append(h)

        #plt.plot((pp[0]+pp[1])/2,h,'o',color=col[c],ms=4,zorder=1000)

        pos[c] = (pp[0]+pp[1])/2,h
        #print "ciao"

    #axarr[0].set_yticks(np.arange(0,1.2,0.2))
    #axarr[0].set_ylim([0,max(hm)+0.05])

    #axarr[0].yaxis.tick_right()
    #axarr[0].tick_params(labelsize=14)
    #axarr[0].set_ylabel('Distance',fontsize=18)
    #axarr[0].yaxis.set_label_position("right")

    if nan==False:
        R = np.corrcoef(x)
    else:
        R =  np.array(pd.DataFrame(x.T).corr()) 

    ims = arr.imshow(R,vmin=-1,vmax=1,cmap=cm.RdBu_r,interpolation='nearest')



    cbar_ax = fig.add_axes(posx)
    cbr = fig.colorbar(ims, cax=cbar_ax)


    cbr.set_label('correlation',fontsize=18)
    arr.set_xlabel(xylabel,fontsize=18)
    arr.set_ylabel(xylabel,fontsize=18)
    arr.tick_params(labelsize=14)

    arr.set_ylim([0,N])

    for l in validated:
        if l == tuple(range(N)): continue

        arr.add_patch(
            patches.Rectangle(
                (min(l)-0.5, min(l)-0.5),
                len(l),
                len(l),
                fill=False,
                lw=1.,
                color='green',	
                )
            )

    plt.setp(arr.get_xticklabels(), visible=False)
    plt.subplots_adjust(hspace=.0)


    if file_w!=None:    
        plt.savefig(file_w+'.pdf',bbox_inches='tight')
    else:
        plt.show()

